/* 
 * GtkGL# - OpenGL Graphics Library for the Gtk# Toolkit
 *  
 * Copyright (c) 2002-2004 The Olympum Group, http://www.olympum.com/
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer. 
 * 
 * - Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 
 * - Neither the name of The Olympum Group nor the names of its contributors 
 * may be used to endorse or promote products derived from this software without 
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
 
namespace Gdk
{
	using System;
	using System.Runtime.InteropServices;
	
	internal class GLContext
	{
		[DllImport("X11")]
		static extern void XFree (IntPtr reference);

		[DllImport("GL")]
		static extern IntPtr glXCreateContext (IntPtr display, IntPtr visualInfo, IntPtr shareList, bool direct);
				
		[DllImport("GL")]
		static extern IntPtr glXChooseVisual (IntPtr display, int screen, int[] attribList);

		[DllImport("gdk-x11-2.0")]
		static extern IntPtr gdk_x11_get_default_xdisplay();
		
		[DllImport("gdk-x11-2.0")]
		static extern int gdk_x11_get_default_screen();
		
		// the Xlib Display used
		IntPtr xdisplay;
		// the Xlib GLXContext used
		IntPtr glxcontext;
		
		internal GLContext (int[] attributeList)
		{			
			// choose the visual based on attribute list  			
			xdisplay = gdk_x11_get_default_xdisplay();
			IntPtr visualInfo = glXChooseVisual (xdisplay, 
				gdk_x11_get_default_screen(), 
				attributeList);
			
			// create glxcontext using visual
			try {
				glxcontext = glXCreateContext (xdisplay, 
					visualInfo, 
					IntPtr.Zero, 
					true);
			} finally {
				XFree (visualInfo);
			}
		}

		[DllImport("GL")]
		public static extern void glXDestroyContext(IntPtr display, IntPtr context);

		~GLContext ()
		{
			glXDestroyContext (xdisplay, glxcontext);
		}

		[DllImport("gdk-x11-2.0")]
		static extern int gdk_x11_drawable_get_xid(IntPtr drawable);

		[DllImport("GL")]
		static extern bool glXMakeCurrent (IntPtr display, int drawableID, IntPtr glxcontext);

		public bool MakeCurrent (Window window)
		{
			int id = gdk_x11_drawable_get_xid(window.Handle);
			return glXMakeCurrent(xdisplay, id, glxcontext);
		}

		[DllImport("GL")]
		static extern void glXSwapBuffers (IntPtr display, int drawableID);

		public void SwapBuffers (Window window)
		{
			int id = gdk_x11_drawable_get_xid(window.Handle);
			glXSwapBuffers(xdisplay, id);		
		}		
	}
}
